/* crypt-gpgme.c - GPGMEbased crypto operations
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * Copyright (C) 1998,1999,2000 Thomas Roessler <roessler@guug.de>
 * Copyright (C) 2001  Thomas Roessler <roessler@guug.de>
 *                     Oliver Ehli <elmy@acm.org>
 * Copyright (C) 2002 g10 Code GmbH
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */


#include "mutt.h"

#ifdef HAVE_GPGME

#include "mutt_curses.h"
#include "crypt.h"
#include "mime.h"
#include "copy.h"

#include <sys/wait.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif

#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
#endif

#ifdef HAVE_SYS_RESOURCE_H
# include <sys/resource.h>
#endif

#include "crypt.h"


static char *find_keys (ADDRESS *to, ADDRESS *cc, ADDRESS *bcc,
                        unsigned int *r_application);
static int verify_one (BODY *sigbdy, STATE *s,
                       const char *tempfile, int is_smime);
static BODY *sign_message (BODY *a, int use_smime);
static BODY *encrypt_message_pgp (BODY *a, char *keylist, int sign);
static BODY *encrypt_message_smime (BODY *a, char *certlist);


/* fixme: duplicated from pgp.c - should get factored out */
int mutt_is_application_pgp (BODY *m)
{
  int t = 0;
  char *p;
  
  if (m->type == TYPEAPPLICATION)
  {
    if (!ascii_strcasecmp (m->subtype, "pgp") || !ascii_strcasecmp (m->subtype, "x-pgp-message"))
    {
      if ((p = mutt_get_parameter ("x-action", m->parameter))
	  && (!ascii_strcasecmp (p, "sign") || !ascii_strcasecmp (p, "signclear")))
	t |= PGPSIGN;

      if ((p = mutt_get_parameter ("format", m->parameter)) && 
	  !ascii_strcasecmp (p, "keys-only"))
	t |= PGPKEY;

      if(!t) t |= PGPENCRYPT;  /* not necessarily correct, but... */
    }

    if (!ascii_strcasecmp (m->subtype, "pgp-signed"))
      t |= PGPSIGN;

    if (!ascii_strcasecmp (m->subtype, "pgp-keys"))
      t |= PGPKEY;
  }
  else if (m->type == TYPETEXT && ascii_strcasecmp ("plain", m->subtype) == 0)
  {
    if (((p = mutt_get_parameter ("x-mutt-action", m->parameter))
	 || (p = mutt_get_parameter ("x-action", m->parameter)) 
	 || (p = mutt_get_parameter ("action", m->parameter)))
	 && !ascii_strncasecmp ("pgp-sign", p, 8))
      t |= PGPSIGN;
    else if (p && !ascii_strncasecmp ("pgp-encrypt", p, 11))
      t |= PGPENCRYPT;
  }
  return t;
}

/* fixme: Duplicated from smime.c */
int mutt_is_application_smime (BODY *m)
{
  char *t=NULL;
  int len, complain=0;

  if ((m->type & TYPEAPPLICATION) && m->subtype)
  {
    /* S/MIME MIME types don't need x- anymore, see RFC2311 */
    if (!ascii_strcasecmp (m->subtype, "x-pkcs7-mime") ||
	!ascii_strcasecmp (m->subtype, "pkcs7-mime"))
    {
      if ((t = mutt_get_parameter ("smime-type", m->parameter)))
      {
	if (!ascii_strcasecmp (t, "enveloped-data"))
	  return SMIMEENCRYPT;
	else if (!ascii_strcasecmp (t, "signed-data"))
	  return (SMIMESIGN|SMIMEOPAQUE);
	else return 0;
      }
      /* Netscape 4.7 uses 
       * Content-Description: S/MIME Encrypted Message
       * instead of Content-Type parameter
       */
      if (!ascii_strcasecmp (m->description, "S/MIME Encrypted Message"))
	return SMIMEENCRYPT;
      complain = 1;
    }
    else if (ascii_strcasecmp (m->subtype, "octet-stream"))
      return 0;

    t = mutt_get_parameter ("name", m->parameter);

    if (!t) t = m->d_filename;
    if (!t) t = m->filename;
    if (!t) 
    {
      if (complain)
	mutt_message (_("S/MIME messages with no hints on content are unsupported."));
      return 0;
    }

    /* no .p7c, .p10 support yet. */

    len = mutt_strlen (t) - 4;
    if (len > 0 && *(t+len) == '.')
    {
      len++;
      if (!ascii_strcasecmp ((t+len), "p7m"))
#if 0
       return SMIMEENCRYPT;
#else
      /* Not sure if this is the correct thing to do, but 
         it's required for compatibility with Outlook */
       return (SMIMESIGN|SMIMEOPAQUE);
#endif
      else if (!ascii_strcasecmp ((t+len), "p7s"))
	return (SMIMESIGN|SMIMEOPAQUE);
    }
  }

  return 0;
}


static void print_time(time_t t, STATE *s)
{
  char p[STRING];

  setlocale (LC_TIME, "");
  strftime (p, sizeof (p), "%c", localtime (&t));
  setlocale (LC_TIME, "C");
  state_attach_puts (p, s);
}

/* print the current time to avoid spoofing of the signature output.
   This is not used with gpgme */
void crypt_current_time(STATE *s, char *app_name)
{
  time_t t;
  char p[STRING], tmp[STRING];

  if (option (OPTCRYPTTIMESTAMP))
  {
    t = time(NULL);
    setlocale (LC_TIME, "");
    strftime (p, sizeof (p), _(" (current time: %c)"), localtime (&t));
    setlocale (LC_TIME, "C");
  }
  else
    *p = '\0';

  snprintf (tmp, sizeof (tmp), _("[-- %s output follows%s --]\n"), NONULL(app_name), p);
  state_attach_puts (tmp, s);
}



void crypt_forget_passphrase (void)
{
  /* fixme: notify gpg-agent to forget the passphrases.  Note that
     this function is not anymoreso important as don't have to care
     about passphrases in Mutt at all */
  mutt_message _("Passphrase(s) are handled by gpg-agent - forget not yet implemented");
}


# if defined(HAVE_SETRLIMIT) && (!defined(DEBUG))

static void disable_coredumps (void)
{
  /* This is a dummy - we don't handle passphrases, so coredumps are okay */
}

# endif /* HAVE_SETRLIMIT */


int crypt_valid_passphrase(int flags)
{
  return 1; /* always valid because it is handled by gpg-agent */
}



int mutt_protect (HEADER *msg, char *keylist)
{
  BODY *pbody = NULL, *tmp_pbody = NULL;
  BODY *tmp_smime_pbody = NULL;
  BODY *tmp_pgp_pbody = NULL;
  int flags = msg->security;

#if 0/* traditional encryption */
  int traditional = 0;
  /* FIXME: take code from crypt.c */
#endif
  
  if (!isendwin ())
    mutt_endwin (NULL);

  tmp_smime_pbody = msg->content;

  if (msg->security & SIGN)
    {
      if (msg->security & APPLICATION_SMIME)
        {
          if (!(tmp_pbody = sign_message (msg->content, 1)))
            return -1;
          pbody = tmp_smime_pbody = tmp_pbody;
        }

      if ((msg->security & APPLICATION_PGP) &&
          (!(flags & ENCRYPT) || option (OPTPGPRETAINABLESIG)))
        {
          if (!(tmp_pbody = sign_message (msg->content, 0)))
            return -1;
          
          flags &= ~SIGN;
          pbody = tmp_pgp_pbody = tmp_pbody;
        }

      if ((msg->security & APPLICATION_SMIME) &&
          (msg->security & APPLICATION_PGP))
        {
          /* here comes the draft ;-) */
        }
    }


  if (msg->security & ENCRYPT)
    {
      if (msg->security & APPLICATION_SMIME)
        {
          if (!(tmp_pbody = encrypt_message_smime (tmp_smime_pbody, keylist)))
            {
              /* signed ? free it! */
              return (-1);
            }
          /* free tmp_body if messages was signed AND encrypted ... */
          if (tmp_smime_pbody != msg->content && tmp_smime_pbody != tmp_pbody)
            {
              /* detatch and dont't delete msg->content,
                 which tmp_smime_pbody->parts after signing. */
              tmp_smime_pbody->parts = tmp_smime_pbody->parts->next;
              msg->content->next = NULL;
              mutt_free_body (&tmp_smime_pbody);
            }
          pbody = tmp_pbody;
        }

      if (msg->security & APPLICATION_PGP)
        {
          if (!(pbody = encrypt_message_pgp (msg->content, keylist,
                                             flags & SIGN)))
            {
              /* did we perform a retainable signature? */
              if (flags != msg->security)
                {
                  /* remove the outer multipart layer */
                  msg->content = mutt_remove_multipart (msg->content);
                  /* get rid of the signature */
                  mutt_free_body (&msg->content->next);
                }
              
              return (-1);
            }

          /* destroy temporary signature envelope when doing
             retainable signatures.  */
          if (flags != msg->security)
            {
              mutt_remove_multipart (msg->content);
              mutt_free_body (&msg->content->next);
            }
        }
    }

  if(pbody)
    msg->content = pbody;
  
  return 0;
}


   
     
int mutt_is_multipart_signed (BODY *b)
{
  char *p;

  if (!b || !(b->type == TYPEMULTIPART) ||
      !b->subtype || mutt_strcasecmp(b->subtype, "signed"))
    return 0;

  if (!(p = mutt_get_parameter("protocol", b->parameter)))
    return 0;

  if (!(mutt_strcasecmp (p, "multipart/mixed")))
    return SIGN;

  if (!(mutt_strcasecmp (p, "application/pgp-signature")))
    return PGPSIGN;
    
  if (!(mutt_strcasecmp(p, "application/x-pkcs7-signature")))
    return SMIMESIGN;
  if (!(mutt_strcasecmp(p, "application/pkcs7-signature")))
    return SMIMESIGN;

  return 0;
}
   
     
int mutt_is_multipart_encrypted (BODY *b)
{
  char *p;
  
  if (!b || b->type != TYPEMULTIPART ||
      !b->subtype || ascii_strcasecmp (b->subtype, "encrypted") ||
      !(p = mutt_get_parameter ("protocol", b->parameter)) ||
      ascii_strcasecmp (p, "application/pgp-encrypted"))
    return 0;
  
  return PGPENCRYPT;
}



/* Check out the type of encryption used and set the cached status
   values if there are any */
int crypt_query (BODY *m)
{
  int t = 0;

  if (m->type == TYPEAPPLICATION)
  {
    t |= mutt_is_application_pgp(m);
    t |= mutt_is_application_smime(m);
    if (t && m->goodsig) 
      t |= GOODSIGN;
    if (t && m->badsig) 
      t |= BADSIGN;
  }
  else if (m->type == TYPETEXT)
  {
    t |= mutt_is_application_pgp (m);
    if (t && m->goodsig)
      t |= GOODSIGN;
  }
  else if (m->type == TYPEMULTIPART)
  {
    t |= mutt_is_multipart_encrypted(m);
    t |= mutt_is_multipart_signed (m);

    if (t && m->goodsig)
      t |= GOODSIGN;
  }

  if (m->type == TYPEMULTIPART || m->type == TYPEMESSAGE)
  {
    BODY *p;
 
    for (p = m->parts; p; p = p->next)
      t |= crypt_query (p) & ~GOODSIGN;
  }

  return t;
}



/* Write the message body/part A described by state S to a the given
   TEMPFILE.  fixme: There is no need to use the tempfiles with gpgme
   but for now we go with this. */
int crypt_write_signed(BODY *a, STATE *s, const char *tempfile)
{
  FILE *fp;
  int c;
  short hadcr;
  size_t bytes;

  if (!(fp = safe_fopen (tempfile, "w")))
  {
    mutt_perror (tempfile);
    return -1;
  }
      
  fseek (s->fpin, a->hdr_offset, 0);
  bytes = a->length + a->offset - a->hdr_offset;
  hadcr = 0;
  while (bytes > 0)
  {
    if ((c = fgetc (s->fpin)) == EOF)
      break;
    
    bytes--;
    
    if  (c == '\r')
      hadcr = 1;
    else 
    {
      if (c == '\n' && !hadcr)
	fputc ('\r', fp);
      
      hadcr = 0;
    }
    
    fputc (c, fp);
    
  }
  fclose (fp);

  return 0;
}


void convert_to_7bit (BODY *a)
{
  while (a)
  {
    if (a->type == TYPEMULTIPART)
    {
      if (a->encoding != ENC7BIT)
      {
        a->encoding = ENC7BIT;
	convert_to_7bit(a->parts);
      }
      else if (option (OPTPGPSTRICTENC))
	convert_to_7bit (a->parts);
    } 
    else if (a->type == TYPEMESSAGE &&
	     mutt_strcasecmp(a->subtype, "delivery-status"))
    {
      if(a->encoding != ENC7BIT)
	mutt_message_to_7bit (a, NULL);
    }
    else if (a->encoding == ENC8BIT)
      a->encoding = ENCQUOTEDPRINTABLE;
    else if (a->encoding == ENCBINARY)
      a->encoding = ENCBASE64;
    else if (a->content && a->encoding != ENCBASE64 &&
	     (a->content->from || (a->content->space && 
				   option (OPTPGPSTRICTENC))))
      a->encoding = ENCQUOTEDPRINTABLE;
    a = a->next;
  }
}



void crypt_extract_keys_from_messages (HEADER * h)
{
#if 0 
  /* FIXME take code from crypt.c */
#endif
}



/* Do a quick check to make sure that we can find all of the
   encryption keys if the user has requested this service. 
   Return the list of keys in KEYLIST. */
int crypt_get_keys (HEADER *msg, char **keylist)
{
  unsigned int app;

  set_option (OPTPGPCHECKTRUST);
  
  *keylist = NULL;
  
  if (msg->security & ENCRYPT)
    {
      if ( (msg->security & (APPLICATION_PGP|APPLICATION_SMIME)))
        {
          if (!(*keylist = find_keys (msg->env->to, msg->env->cc,
                                      msg->env->bcc, &app)))
            return -1;
          if (!app)
            {
              safe_free ((void**)keylist);
              *keylist = NULL;
              mutt_error (_("Error: can't mix key types"));
              return -1;
            }
          unset_option (OPTPGPCHECKTRUST);
          msg->security &= ~(APPLICATION_PGP|APPLICATION_SMIME);
          if (app & APPLICATION_PGP)
            msg->security |= APPLICATION_PGP;
          if (app & APPLICATION_SMIME)
            msg->security |= APPLICATION_SMIME;
        }
    }
    
  return 0;
}



static void crypt_fetch_signatures (BODY ***signatures, BODY *a, int *n)
{
  for (; a; a = a->next)
  {
    if (a->type == TYPEMULTIPART)
      crypt_fetch_signatures (signatures, a->parts, n);
    else
    {
      if((*n % 5) == 0)
	safe_realloc ((void **) signatures, (*n + 6) * sizeof (BODY **));

      (*signatures)[(*n)++] = a;
    }
  }
}


/* This routine verifies a "multipart/signed" body.  */
void mutt_signed_handler (BODY *a, STATE *s)
{
  char tempfile[_POSIX_PATH_MAX]; 
  char *protocol;
  int protocol_major = TYPEOTHER;
  char *protocol_minor = NULL;
  
  BODY *b = a;
  BODY **signatures = NULL;
  int sigcnt = 0;
  int i;
  short goodsig = 1;

  protocol = mutt_get_parameter ("protocol", a->parameter);
  a = a->parts;

  /* extract the protocol information */
  
  if (protocol)
  {
    char major[STRING];
    char *t;

    if ((protocol_minor = strchr (protocol, '/')))
      protocol_minor++;
    
    strfcpy (major, protocol, sizeof(major));
    if((t = strchr(major, '/')))
      *t = '\0';
    
    protocol_major = mutt_check_mime_type (major);
  }

  /* consistency check */
  if (!(a && a->next && a->next->type == protocol_major && 
      !mutt_strcasecmp (a->next->subtype, protocol_minor)))
  {
    state_attach_puts (_("[-- Error: Inconsistent multipart/signed structure! --]\n\n"), s);
    mutt_body_handler (a, s);
    return;
  }

  
  if (protocol_major == TYPEAPPLICATION &&
      !mutt_strcasecmp (protocol_minor, "pgp-signature"))
    ;
  else if (protocol_major == TYPEAPPLICATION &&
           !mutt_strcasecmp (protocol_minor, "x-pkcs7-signature"))
    ; 
  else if (protocol_major == TYPEAPPLICATION &&
           !mutt_strcasecmp (protocol_minor, "pkcs7-signature"))
    ; 
  else if (protocol_major == TYPEMULTIPART &&
           !mutt_strcasecmp (protocol_minor, "mixed"))
    ;
  else
    {
      state_printf (s, _("[-- Error: Unknown multipart/signed "
                         "protocol %s! --]\n\n"), protocol);
      mutt_body_handler (a, s);
      return;
    }
  
  if (s->flags & M_DISPLAY)
    {
      crypt_fetch_signatures (&signatures, a->next, &sigcnt);
      
      if (sigcnt)
        {
          mutt_mktemp (tempfile);
          if (crypt_write_signed (a, s, tempfile) == 0)
            {
              for (i = 0; i < sigcnt; i++)
                {
                  if (signatures[i]->type == TYPEAPPLICATION 
                      && !mutt_strcasecmp (signatures[i]->subtype,
                                           "pgp-signature"))
                    {
                      if (verify_one (signatures[i], s, tempfile, 0))
                        goodsig = 0;
                      continue;
                    }
                  if (signatures[i]->type == TYPEAPPLICATION 
                      && !mutt_strcasecmp (signatures[i]->subtype,
                                           "pkcs7-signature"))
                    {
                      if (verify_one (signatures[i], s, tempfile, 1))
                        goodsig = 0;
                      continue;
                    }
                  if (signatures[i]->type == TYPEAPPLICATION 
                      && !mutt_strcasecmp (signatures[i]->subtype,
                                           "x-pkcs7-signature"))
                    {
                      if (verify_one (signatures[i], s, tempfile, 1))
                        goodsig = 0;
                      continue;
                    }
                  state_printf (s, _("[-- Warning: We can't verify "
                                     "%s/%s signatures. --]\n\n"),
                                TYPE(signatures[i]), signatures[i]->subtype);
                }
            }
          
          mutt_unlink (tempfile);

          b->goodsig = goodsig;
          b->badsig = goodsig; /* FIXME: strange usage */
      
          /* Now display the signed body */
          state_attach_puts (_("[-- The following data is signed --]\n\n"), s);
          safe_free((void **) &signatures);
        }
      else
        state_attach_puts (_("[-- Warning: Can't find "
                             "any signatures. --]\n\n"), s);
    }
  
  mutt_body_handler (a, s);
  
  if (s->flags & M_DISPLAY && sigcnt)
    state_attach_puts (_("\n[-- End of signed data --]\n"), s);
}



/* This routine attempts to find the keyids of the recipients of a
   message.  It returns NULL if any of the keys can not be found.  */
static char *find_keys (ADDRESS *to, ADDRESS *cc, ADDRESS *bcc,
                        unsigned int *r_application)
{
  char *keyID, *keylist = NULL, *t;
  size_t keylist_size = 0;
  size_t keylist_used = 0;
  ADDRESS *tmp = NULL, *addr = NULL;
  ADDRESS **last = &tmp;
  ADDRESS *p, *q;
  int i;
  crypt_key_t *k_info, *key;
  const char *fqdn = mutt_fqdn (1);

  *r_application = APPLICATION_PGP|APPLICATION_SMIME;
  
  for (i = 0; i < 3; i++) 
    {
      switch (i)
        {
        case 0: p = to; break;
        case 1: p = cc; break;
        case 2: p = bcc; break;
        default: abort ();
        }
      
      *last = rfc822_cpy_adr (p);
      while (*last)
        last = &((*last)->next);
    }
  
  if (fqdn)
    rfc822_qualify (tmp, fqdn);
  
  tmp = mutt_remove_duplicates (tmp);
  
  for (p = tmp; p ; p = p->next)
    {
      char buf[LONG_STRING];
      
      q = p;
      k_info = NULL;
      
      if ((keyID = mutt_crypt_hook (p)) != NULL)
        {
          int r;
          snprintf (buf, sizeof (buf), _("Use keyID = \"%s\" for %s?"),
                    keyID, p->mailbox);
          if ((r = mutt_yesorno (buf, M_YES)) == M_YES)
            {
              /* check for e-mail address */
              if ((t = strchr (keyID, '@')) && 
                  (addr = rfc822_parse_adrlist (NULL, keyID)))
                {
                  if (fqdn)
                    rfc822_qualify (addr, fqdn);
                  q = addr;
                }
              else
                k_info = crypt_getkeybystr (keyID, KEYFLAG_CANENCRYPT, 
                                            *r_application);
            }
          else if (r == -1)
            {
              safe_free ((void **) &keylist);
              rfc822_free_address (&tmp);
              rfc822_free_address (&addr);
              return NULL;
            }
        }

#if 0 /* FIXME: for what is this good? */      
      if (k_info == NULL)
        crypt_invoke_getkeys (q);
#endif

      if (k_info == NULL
          && (k_info = crypt_getkeybyaddr (q, KEYFLAG_CANENCRYPT)) == NULL)
        {
          snprintf (buf, sizeof (buf), _("Enter keyID for %s: "), q->mailbox);
          
          if ((key = crypt_ask_for_key (buf, q->mailbox,
                                        KEYFLAG_CANENCRYPT,
                                        *r_application)) == NULL)
            {
              safe_free ((void **)&keylist);
              rfc822_free_address (&tmp);
              rfc822_free_address (&addr);
              return NULL;
            }
        }
      else
        key = k_info;

      {
        const char *s = crypt_fpr (key);

        if (key->flags & KEYFLAG_ISX509)
          *r_application &= ~APPLICATION_PGP;
        if (!(key->flags & KEYFLAG_ISX509))
          *r_application &= ~APPLICATION_SMIME;
      
        keylist_size += mutt_strlen (s) + 4;
        safe_realloc ((void **)&keylist, keylist_size);
        sprintf (keylist + keylist_used, "%s0x%s",
                 keylist_used ? " " : "", /* __SPRINTF_CHECKED__ */ s);
      }
      keylist_used = mutt_strlen (keylist);
        
      crypt_free_key (&key);
      rfc822_free_address (&addr);
    }
  rfc822_free_address (&tmp);
  return (keylist);
}



static GpgmeData body_to_data_object (BODY *a, int convert)
{
  char tempfile[_POSIX_PATH_MAX];
  FILE *fptmp;
  int err = 0;
  GpgmeData data;
  
  mutt_mktemp (tempfile);
  fptmp = safe_fopen (tempfile, "w+");
  if (!fptmp)
    {
      mutt_perror (tempfile);
      return NULL;
    }

  mutt_write_mime_header (a, fptmp);
  fputc ('\n', fptmp);
  mutt_write_mime_body (a, fptmp);

  if (convert)
    {
      err = gpgme_data_new (&data);
      if (!err)
        {
          int c, hadcr = 0;
          unsigned char buf[1];
          
          rewind (fptmp);
          while ((c = fgetc (fptmp)) != EOF)
            {
              if  (c == '\r')
                hadcr = 1;
              else 
                {
                  if (c == '\n' && !hadcr)
                    {
                      buf[0] = '\r';
                      gpgme_data_write (data, buf, 1);
                    }
                  
                  hadcr = 0;
                }
              /* FIXME: This is quite suboptimal */
              buf[0] = c;
              gpgme_data_write (data, buf, 1);
            }
        }
      fclose(fptmp);
    }
  else
    {
      fclose(fptmp);
      err = gpgme_data_new_from_file (&data, tempfile, 1);
    }
  unlink (tempfile);
  if (err) 
    {
      mutt_error ("error allocating data object: %s\n", gpgme_strerror (err));
      return NULL;
    }

  return data;
}


static GpgmeData file_to_data_object (FILE *fp, long offset, long length)
{
  int err = 0;
  GpgmeData data;
  
  err = gpgme_data_new_from_filepart (&data, NULL, fp, offset, length);
  if (err) 
    {
      mutt_error ("error allocating data object: %s\n", gpgme_strerror (err));
      return NULL;
    }

  return data;
}

/* Write a GPGME data object to the stream FP */
static int data_object_to_stream (GpgmeData data, FILE *fp)
{
  int err;
  char buf[4096], *p;
  size_t nread;

  err = gpgme_data_rewind (data);
  if (err)
    {
      mutt_error ("error rewinding data object: %s\n", gpgme_strerror (err));
      return -1;
    }

  while ( !(err = gpgme_data_read (data, buf, sizeof buf, &nread)) )
    {
      /* fixme: we are not really converting CRLF to LF but just
         skipping CR. Doinh it correctly needs a more complex logic */
      for (p=buf; nread; p++, nread--)
        {
          if (*p != '\r')
            putc (*p, fp);
        }

     if (ferror (fp))
       {
         mutt_perror ("[tempfile]");
         return -1;
       }
    }
  if (err != GPGME_EOF)
    {
      mutt_error ("error reading data object: %s\n", gpgme_strerror (err));
      return -1;
    }
  return 0;
}

/* Copy a data object to a newly created temporay file and return that
   filename. Caller must free. */
static char *data_object_to_tempfile (GpgmeData data)
{
  int err;
  char tempfile[_POSIX_PATH_MAX];
  FILE *fp;

  mutt_mktemp (tempfile);
  fp = safe_fopen (tempfile, "w+");
  if (!fp)
    {
      mutt_perror (tempfile);
      return NULL;
    }

  err = gpgme_data_rewind (data);
  if (!err)
    {
      char buf[4096];
      size_t nread;

      while ( !(err = gpgme_data_read (data, buf, sizeof buf, &nread)) )
        {
          if (fwrite (buf, nread, 1, fp) != 1)
            {
              mutt_perror (tempfile);
              fclose (fp);
              unlink (tempfile);
              return NULL;
            }
        }
    }
  fclose (fp);
  if (err != GPGME_EOF) 
    {
      mutt_error ("error reading data object: %s\n", gpgme_strerror (err));
      unlink (tempfile);
      return NULL;
    }
  
  return safe_strdup (tempfile);
}


static GpgmeRecipients create_recipient_set (const char *keylist)
{
  int err;
  const char *s;
  char buf[100];
  int i;
  GpgmeRecipients rset;

  err = gpgme_recipients_new (&rset);
  if (err)
    {
      mutt_error ("error creating recipient set: %s\n", gpgme_strerror (err));
      return NULL;
    }
  
  s = keylist;
  do {
    while (*s == ' ')
      s++;
    for (i=0; *s && *s != ' ' && i < sizeof(buf)-1;)
      buf[i++] = *s++;
    buf[i] = 0;
    if (*buf)
      {
        err = gpgme_recipients_add_name (rset, buf);
        if (err)
          {
            mutt_error ("error setting recipient `%s': %s\n",
                        buf, gpgme_strerror (err));
            gpgme_recipients_release (rset);
            return NULL;
          }
      }
  } while (*s);

  return rset;
}


/* Encrypt the gpgme data object PLAINTEXT to the recipients in RSET
   and return an allocated filename to a temporary file containing the
   enciphered text.  With USE_SMIME set to true, the smime backend is
   used.  Return NULL in case of error */
static char *encrypt_gpgme_object (GpgmeData plaintext,
                                   GpgmeRecipients rset,
                                   int use_smime)
{
  int err;
  GpgmeCtx ctx;
  GpgmeData ciphertext;
  char *outfile;

  err = gpgme_new (&ctx);
  if (err)
    {
      mutt_error ("error creating gpgme context: %s\n", gpgme_strerror (err));
      return NULL;
    }
  if (use_smime) 
    {
      err = gpgme_set_protocol (ctx, GPGME_PROTOCOL_CMS);
      if (err)
        {
          mutt_error ("error enabling CMS protocol: %s\n",
                      gpgme_strerror (err));
          gpgme_release (ctx);
          return NULL;
        }
    }
  else
    {
      gpgme_set_armor (ctx, 1);
    }
  err = gpgme_data_new (&ciphertext);
  if (err) 
    {
      mutt_error ("error allocating ciphertext data object: %s\n",
                  gpgme_strerror (err));
      gpgme_release (ctx);
      return NULL;
    }

  err = gpgme_op_encrypt (ctx, rset, plaintext, ciphertext);
  if (err)
    {
      mutt_error ("error encrypting data: %s\n", gpgme_strerror (err));
      gpgme_data_release (ciphertext);
      gpgme_release (ctx);
      return NULL;
    }

  gpgme_release (ctx);

  outfile = data_object_to_tempfile (ciphertext);
  gpgme_data_release (ciphertext);
  return outfile;
}


static int get_micalg (GpgmeCtx ctx, char *buf, size_t buflen)
{
  int i;
  const char *s = gpgme_get_op_info (ctx, 0);

  if (!buflen)
    return -1;
  *buf = 0;
  if (s && (s = strstr (s, "<micalg>")))
    {
      s += 8;
      for (i=0; *s && *s != '<' && *s != '>' && i < buflen-1;)
        buf[i++] = *s++;
      buf[i] = 0;
    }
  return *buf? 0:-1;
}


static BODY *sign_message (BODY *a, int use_smime)
{
  BODY *t;
  char *sigfile;
  int err = 0;
  GpgmeCtx ctx;
  GpgmeData message, signature;

  convert_to_7bit (a); /* Signed data _must_ be in 7-bit format. */

  message = body_to_data_object (a, 1);
  if (!message)
     return NULL;
  err = gpgme_data_new (&signature);
  if (err) 
    {
      gpgme_data_release (message);
      mutt_error ("error allocating new data object: %s\n",
                  gpgme_strerror (err));
      return NULL;
    }

  err = gpgme_new (&ctx);
  if (err)
    {
      gpgme_data_release (signature);
      gpgme_data_release (message);
      mutt_error ("error creating gpgme context: %s\n", gpgme_strerror (err));
      return NULL;
    }
  if (use_smime)
    gpgme_set_protocol (ctx, GPGME_PROTOCOL_CMS);
  else
    gpgme_set_armor (ctx, 1);

  err = gpgme_op_sign (ctx, message, signature, GPGME_SIG_MODE_DETACH );
  gpgme_data_release (message);
  if (err)
    {
      gpgme_data_release (signature);
      gpgme_release (ctx);
      mutt_error ("error signing data: %s\n", gpgme_strerror (err));
      return NULL;
    }

  sigfile = data_object_to_tempfile (signature);
  gpgme_data_release (signature);
  if (!sigfile)
    {
      gpgme_release (ctx);
      return NULL;
    }

  t = mutt_new_body ();
  t->type = TYPEMULTIPART;
  t->subtype = safe_strdup ("signed");
  t->encoding = ENC7BIT;
  t->use_disp = 0;
  t->disposition = DISPINLINE;

  mutt_generate_boundary (&t->parameter);
  if (use_smime)
    {
      mutt_set_parameter ("protocol", "application/pkcs7-signature",
                          &t->parameter);
      /* FIXME: Extend gpgme to return a micalg also for S/MIME */
      mutt_set_parameter ("micalg", "sha1", &t->parameter);
    }
  else 
    {
      char buf[100];
      mutt_set_parameter ("protocol", "application/pgp-signature",
                          &t->parameter);
      if (!get_micalg (ctx, buf, sizeof buf))
        mutt_set_parameter ("micalg", buf, &t->parameter);
    }
  gpgme_release (ctx);

  t->parts = a;
  a = t;

  t->parts->next = mutt_new_body ();
  t = t->parts->next;
  t->type = TYPEAPPLICATION;
  if (use_smime)
    {
      t->subtype = safe_strdup ("pkcs7-signature");
      mutt_set_parameter ("name", "smime.p7s", &t->parameter);
      t->encoding = ENCBASE64; 
      t->use_disp = 1;
      t->disposition = DISPATTACH;
      t->d_filename = safe_strdup ("smime.p7s");
    }
  else
    {
      t->subtype = safe_strdup ("pgp-signature");
      t->use_disp = 0;
      t->disposition = DISPINLINE;
      t->encoding = ENC7BIT;
    }
  t->filename = sigfile;
  t->unlink = 1; /* ok to remove this file after sending. */

  return a;
}


/* Encrypt the mail body A to all keys given as space separated keyids
   or fingerprints in KEYLIST and return the encrypted body.  */
static BODY *encrypt_message_pgp (BODY *a, char *keylist, int sign)
{
  char *outfile = NULL;
  BODY *t;
  GpgmeRecipients rset;
  GpgmeData plaintext;
  
  rset = create_recipient_set (keylist);
  if (!rset)
    return NULL;
  
  if (sign)
    convert_to_7bit (a);
  plaintext = body_to_data_object (a, 0);
  if (!plaintext)
    {
      gpgme_recipients_release (rset);
      return NULL;
    }
  
  outfile = encrypt_gpgme_object (plaintext, rset, 0);
  gpgme_data_release (plaintext);
  gpgme_recipients_release (rset);
  if (!outfile)
      return NULL;

  t = mutt_new_body ();
  t->type = TYPEMULTIPART;
  t->subtype = safe_strdup ("encrypted");
  t->encoding = ENC7BIT;
  t->use_disp = 0;
  t->disposition = DISPINLINE;

  mutt_generate_boundary(&t->parameter);
  mutt_set_parameter("protocol", "application/pgp-encrypted", &t->parameter);
  
  t->parts = mutt_new_body ();
  t->parts->type = TYPEAPPLICATION;
  t->parts->subtype = safe_strdup ("pgp-encrypted");
  t->parts->encoding = ENC7BIT;

  t->parts->next = mutt_new_body ();
  t->parts->next->type = TYPEAPPLICATION;
  t->parts->next->subtype = safe_strdup ("octet-stream");
  t->parts->next->encoding = ENC7BIT;
  t->parts->next->filename = outfile;
  t->parts->next->use_disp = 1;
  t->parts->next->disposition = DISPINLINE;
  t->parts->next->unlink = 1; /* delete after sending the message */
  t->parts->next->d_filename = safe_strdup ("msg.asc"); /* non pgp/mime can save */

  return t;
}



static BODY *encrypt_message_smime (BODY *a, char *keylist)
{
  char *outfile = NULL;
  BODY *t;
  GpgmeRecipients rset;
  GpgmeData plaintext;

  rset = create_recipient_set (keylist);
  if (!rset)
    return NULL;

  plaintext = body_to_data_object (a, 0);
  if (!plaintext)
    {
      gpgme_recipients_release (rset);
      return NULL;
    }

  outfile = encrypt_gpgme_object (plaintext, rset, 1);
  gpgme_data_release (plaintext);
  gpgme_recipients_release (rset);
  if (!outfile) 
      return NULL;

  t = mutt_new_body ();
  t->type = TYPEAPPLICATION;
  t->subtype = safe_strdup ("pkcs7-mime");
  mutt_set_parameter ("name", "smime.p7m", &t->parameter);
  mutt_set_parameter ("smime-type", "enveloped-data", &t->parameter);
  t->encoding = ENCBASE64;  /* The output of OpenSSL SHOULD be binary */
  t->use_disp = 1;
  t->disposition = DISPATTACH;
  t->d_filename = safe_strdup ("smime.p7m");
  t->filename = outfile;
  t->unlink = 1; /*delete after sending the message */
  t->parts=0;
  t->next=0;
  
  return t;
}



/* 
 * Copy a clearsigned message, and strip the signature and PGP's
 * dash-escaping.
 * 
 * XXX - charset handling: We assume that it is safe to do
 * character set decoding first, dash decoding second here, while
 * we do it the other way around in the main handler.
 * 
 * (Note that we aren't worse than Outlook &c in this, and also
 * note that we can successfully handle anything produced by any
 * existing versions of mutt.) 
 */

static void copy_clearsigned (FILE *fpin, STATE *s, char *charset)
{
  char buf[HUGE_STRING];
  short complete, armor_header;
  
  FGETCONV *fc;
  
  rewind (fpin);
  
  fc = fgetconv_open (fpin, charset, Charset, M_ICONV_HOOK_FROM);
  
  for (complete = 1, armor_header = 1;
       fgetconvs (buf, sizeof (buf), fc) != NULL;
       complete = strchr (buf, '\n') != NULL)
  {
    if (!complete)
    {
      if (!armor_header)
	state_puts (buf, s);
      continue;
    }

    if (mutt_strcmp (buf, "-----BEGIN PGP SIGNATURE-----\n") == 0)
      break;
    
    if (armor_header)
    {
      if (buf[0] == '\n') 
	armor_header = 0;
      continue;
    }
    
    if (s->prefix) 
      state_puts (s->prefix, s);
    
    if (buf[0] == '-' && buf[1] == ' ')
      state_puts (buf + 2, s);
    else
      state_puts (buf, s);
  }
  
  fgetconv_close (&fc);
}



static const char *
sig_status_string (GpgmeSigStat status)
{
  const char *s = "?";

  switch (status)
    {
    case GPGME_SIG_STAT_NONE:
      s = _("None");
      break;
    case GPGME_SIG_STAT_NOSIG:
      s = _("No Signature");
      break;
    case GPGME_SIG_STAT_GOOD:
    case GPGME_SIG_STAT_GOOD_EXP:
    case GPGME_SIG_STAT_GOOD_EXPKEY:
      s = _("Good");
      break;
    case GPGME_SIG_STAT_BAD:
      s = _("Bad");
      break;
    case GPGME_SIG_STAT_NOKEY:
      s = _("No Key");
      break;
    case GPGME_SIG_STAT_ERROR:
      s = _("Error");
      break;
    case GPGME_SIG_STAT_DIFF:
      s = _("More than one signature");
      break;
    }
  return s;
}


/* Do the actual verification step. With IS_SMIME set to true we
   assume S/MIME (surprise!) */
static int verify_one (BODY *sigbdy, STATE *s,
                       const char *tempfile, int is_smime)
{
  int badsig = -1;
  int err;
  GpgmeCtx ctx;
  GpgmeData signature, message;
  GpgmeSigStat status;

  signature = file_to_data_object (s->fpin, sigbdy->offset, sigbdy->length);
  if (!signature)
    return -1;
  /* We need to tell gpgme about the encoind because the backend can't
     auto-detect plain base-64 encoding whcih is used by S/MIME. */
  if (is_smime)
    gpgme_data_set_encoding (signature, GPGME_DATA_ENCODING_BASE64);

  err = gpgme_data_new_from_file (&message, tempfile, 1);
  if (err) 
    {
      gpgme_data_release (signature);
      mutt_error ("error allocating data object: %s\n", gpgme_strerror (err));
      return -1;
    }
  err = gpgme_new (&ctx);
  if (err)
    {
      mutt_error ("error allocating gpgme context: %s\n",gpgme_strerror (err));
      gpgme_data_release (signature);
      gpgme_data_release (message);
      return -1;
    }
  if (is_smime)
    gpgme_set_protocol (ctx, GPGME_PROTOCOL_CMS);

  /* Note: We don't need a current time output because GPGME avoids
     such an attack by separating the meta information from the
     data. */
  state_attach_puts (_("[-- Begin signature information --]\n"), s);

  err = gpgme_op_verify (ctx, signature, message, &status );
  if (err)
    {
      char buf[200];
      
      snprintf (buf, sizeof(buf)-1, 
                _("Error: verification failed: %s\n"),
                gpgme_strerror (err));
      state_attach_puts (buf, s);
    }
  else
    {
      const char *fpr, *uid, *txt;
      time_t created;
      int idx;
      GpgmeKey key;
      int anybad = 0;

      for(idx=0; (fpr=gpgme_get_sig_status (ctx, idx, &status, &created)); idx++)
        {
          if (status != GPGME_SIG_STAT_GOOD)
            anybad = 1;

          key = NULL;
          if ( !gpgme_get_sig_key (ctx, idx, &key) )
            uid = gpgme_key_get_string_attr (key, GPGME_ATTR_USERID, NULL, 0);
          else
            uid = "[?]";

          if (status == GPGME_SIG_STAT_GOOD
              || status == GPGME_SIG_STAT_GOOD_EXP
              || status == GPGME_SIG_STAT_GOOD_EXPKEY)
            {
              state_attach_puts (_("Good signature from: "), s);
              state_attach_puts (uid, s);
              state_attach_puts ("\n", s);
              state_attach_puts (_("            created: "), s);
              print_time (created, s);
              state_attach_puts ("\n", s);
              if (status == GPGME_SIG_STAT_GOOD_EXP)
                {
                  state_attach_puts (_("Warning: The signature expired at: "),s);
                  print_time ( gpgme_get_sig_ulong_attr
                               (ctx, idx, GPGME_ATTR_EXPIRE, 0), s);
                  state_attach_puts ("\n", s);
                }
              else if (status == GPGME_SIG_STAT_GOOD_EXPKEY)
                {
                  state_attach_puts (_("Warning: The key used to create the "
                                       "signature expired at: "), s);
                  print_time ( gpgme_key_get_ulong_attr
                               (key, GPGME_ATTR_EXPIRE, NULL, 0), s);
                  state_attach_puts ("\n", s);
                }
              txt = _("INTERNAL ERROR: Unknown key validity\n");
              switch ( gpgme_get_sig_ulong_attr(ctx, idx,
                                                GPGME_ATTR_VALIDITY, 0))
                {
                case GPGME_VALIDITY_UNKNOWN:
                  txt = _("Warning: We have NO indication whether "
                          " the key belongs to the person named "
                          "as shown above\n");
                  break;
                case GPGME_VALIDITY_UNDEFINED:
                  break;
                case GPGME_VALIDITY_NEVER:
                  txt = _("WARNING: The key does NOT BELONG to "
                          "the person named as shown above\n");
                  break;
                case GPGME_VALIDITY_MARGINAL:
                  txt = _("Warning: It is NOT certain that the key"
                          " belongs to the person named as shown"
                          " above\n");
                  break;
                case GPGME_VALIDITY_FULL:
                  txt = NULL;
                  break;
                case GPGME_VALIDITY_ULTIMATE:
                  break;
                }
              if (txt)
                state_attach_puts (txt, s);
            }
          else if (status == GPGME_SIG_STAT_BAD)
            {
              state_attach_puts (_("*BAD* signature from: "), s);
              state_attach_puts (uid, s);
              state_attach_puts ("\n\n", s);
            }
          else 
            {
              state_attach_puts (_("Error checking signature: "), s);
              state_attach_puts (sig_status_string (status), s);
              state_attach_puts ("\n\n", s);
            }
          gpgme_key_release (key);
        }
      if (!anybad)
        badsig = 0;
    }

  if (!badsig)
    {
      const char *nota;
      
      if ( (nota=gpgme_get_notation (ctx)) )
        {
          state_attach_puts ("*** Begin Notation ***\n", s);
          state_attach_puts (nota, s);
          if (*nota && nota[strlen (nota)-1] != '\n')
            state_attach_puts ("\n", s);
          state_attach_puts ("*** End Notation ***\n", s);
        }
    }

  gpgme_release (ctx);
  
  state_attach_puts (_("[-- End signature information --]\n\n"), s);
  dprint (1, (debugfile, "pgp_verify_one: returning %d.\n", badsig));
  
  return badsig;
}



static BODY *pgp_decrypt_part (BODY *a, STATE *s, FILE *fpout, BODY *p)
{
  struct stat info;
  BODY *tattach;
  int err;
  GpgmeCtx ctx;
  GpgmeData ciphertext, plaintext;

  /* Make a data object from the body, create context etc. */
  ciphertext = file_to_data_object (s->fpin, a->offset, a->length);
  if (!ciphertext)
    return NULL;
  err = gpgme_data_new (&plaintext);
  if (err)
    {
      mutt_error ("error allocating new data object: %s\n",
                  gpgme_strerror (err));
      gpgme_data_release (ciphertext);
      return NULL;
    }
  err = gpgme_new (&ctx);
  if (err)
    {
      mutt_error ("error allocating gpgme context: %s\n",gpgme_strerror (err));
      gpgme_data_release (ciphertext);
      gpgme_data_release (plaintext);
      return NULL;
    }

  /* Do the decryption */
  err = gpgme_op_decrypt (ctx, ciphertext, plaintext);
  gpgme_data_release (ciphertext);
  gpgme_release (ctx); ctx = NULL;
  if (err)
    {
      if (s->flags & M_DISPLAY)
        {
          char buf[200];
          
          snprintf (buf, sizeof(buf)-1, 
                    _("[-- Error: decryption failed: %s --]\n\n"),
                    gpgme_strerror (err));
          state_attach_puts (buf, s);
        }
      gpgme_data_release (plaintext);
      gpgme_release (ctx);
      return NULL;
  }

  /* Read the output from PGP, and make sure to change CRLF to LF,
     otherwise read_mime_header has a hard time parsing the message.  */
  if (data_object_to_stream (plaintext, fpout))
    {
      gpgme_data_release (plaintext);
      gpgme_release (ctx);
      return NULL;
    }
  gpgme_data_release (plaintext);

  if (s->flags & M_DISPLAY)
  {
/* FIXME     if (!pgp_copy_checksig (pgperr, s->fpout) && p) */
/*        p->goodsig = 1; */
    state_attach_puts (_("[-- End of PGP output --]\n\n"), s);
  }
  gpgme_release (ctx); ctx = NULL;

  fflush (fpout);
  rewind (fpout);
  tattach = mutt_read_mime_header (fpout, 0);
  if (tattach)
    {
      /*
       * Need to set the length of this body part.
       */
      fstat (fileno (fpout), &info);
      tattach->length = info.st_size - tattach->offset;
      
      /* See if we need to recurse on this MIME part.  */
      mutt_parse_part (fpout, tattach);
    }

  return tattach;
}



/* Support for the Application/PGP Content Type. (gpgme version) */
void pgp_application_pgp_handler (BODY *m, STATE *s)
{
  int needpass = -1, pgp_keyblock = 0;
  int clearsign = 0, rv, rc;
  long start_pos = 0;
  long bytes, last_pos, offset;
  char buf[HUGE_STRING];
  char outfile[_POSIX_PATH_MAX];
  char tmpfname[_POSIX_PATH_MAX];
  FILE *pgpout = NULL, *pgpin = NULL, *pgperr = NULL;
  FILE *tmpfp;

  short maybe_goodsig = 1;
  short have_any_sigs = 0;

  char body_charset[STRING];
  mutt_get_body_charset (body_charset, sizeof (body_charset), m);
  
  fseek (s->fpin, m->offset, 0);
  last_pos = m->offset;
  
  for (bytes = m->length; bytes > 0;)
  {
    if (fgets (buf, sizeof (buf), s->fpin) == NULL)
      break;
    
    offset = ftell (s->fpin);
    bytes -= (offset - last_pos); /* don't rely on mutt_strlen(buf) */
    last_pos = offset;
    
    if (mutt_strncmp ("-----BEGIN PGP ", buf, 15) == 0)
    {
      clearsign = 0;
      start_pos = last_pos;

      if (mutt_strcmp ("MESSAGE-----\n", buf + 15) == 0)
        needpass = 1;
      else if (mutt_strcmp ("SIGNED MESSAGE-----\n", buf + 15) == 0)
      {
	clearsign = 1;
        needpass = 0;
      }
      else if (!option (OPTDONTHANDLEPGPKEYS) &&
	       mutt_strcmp ("PUBLIC KEY BLOCK-----\n", buf + 15) == 0)
      {
        needpass = 0;
        pgp_keyblock =1;
      } 
      else
      {
	/* XXX - we may wish to recode here */
	if (s->prefix)
	  state_puts (s->prefix, s);
	state_puts (buf, s);
	continue;
      }

      have_any_sigs = have_any_sigs || (clearsign && (s->flags & M_VERIFY));

      /* Copy PGP material to temporary file */
      mutt_mktemp (tmpfname);
      if ((tmpfp = safe_fopen (tmpfname, "w+")) == NULL)
      {
	mutt_perror (tmpfname);
	return;
      }
      
      fputs (buf, tmpfp);
      while (bytes > 0 && fgets (buf, sizeof (buf) - 1, s->fpin) != NULL)
      {
	offset = ftell (s->fpin);
	bytes -= (offset - last_pos); /* don't rely on mutt_strlen(buf) */
	last_pos = offset;
	
	fputs (buf, tmpfp);

	if ((needpass && mutt_strcmp ("-----END PGP MESSAGE-----\n", buf) == 0) ||
	    (!needpass 
             && (mutt_strcmp ("-----END PGP SIGNATURE-----\n", buf) == 0
                 || mutt_strcmp ("-----END PGP PUBLIC KEY BLOCK-----\n",buf) == 0)))
	  break;
      }

      /* leave tmpfp open in case we still need it - but flush it! */
      fflush (tmpfp);
      
      
      /* Invoke PGP if needed */
      if (!clearsign || (s->flags & M_VERIFY))
      {
	mutt_mktemp (outfile);
	if ((pgpout = safe_fopen (outfile, "w+")) == NULL)
	{
	  mutt_perror (tmpfname);
	  return;
	}
	
/*  	if ((thepid = pgp_invoke_decode (&pgpin, NULL, &pgperr, -1, */
/*  					 fileno (pgpout), -1, tmpfname, */
/*  					 needpass)) == -1) */
/*  	{ */
/*  	  safe_fclose (&pgpout); */
/*  	  maybe_goodsig = 0; */
/*  	  pgpin = NULL; */
/*  	  pgperr = NULL; */
/*  	  state_attach_puts (_("[-- Error: unable to create PGP subprocess! --]\n"), s); */
/*  	} */
/*  	else */ /* PGP started successfully */ 
	{
	  safe_fclose (&pgpin);

	  rv = 0; /*mutt_wait_filter (thepid);*/

	  if (s->flags & M_DISPLAY)
	  {
            state_attach_puts (_("[-- Begin signature information --]\n"), s);
	    rc = -1/*FIXME: pgp_copy_checksig (pgperr, s->fpout)*/;
	    if (rc == 0) have_any_sigs = 1;
	    if (rc || rv) maybe_goodsig = 0;
	  }
	  
	  safe_fclose (&pgperr);
	  
	  if (s->flags & M_DISPLAY)
	  {
	    state_putc ('\n', s);
	    state_attach_puts (_("[-- End signature information --]\n\n"), s);
	  }
	}
      }
      

      /*
       * Now, copy cleartext to the screen.  NOTE - we expect that PGP
       * outputs utf-8 cleartext.  This may not always be true, but it 
       * seems to be a reasonable guess.
       */

      if(s->flags & M_DISPLAY)
      {
	if (needpass)
	  state_attach_puts (_("[-- BEGIN PGP MESSAGE --]\n\n"), s);
	else if (pgp_keyblock)
	  state_attach_puts (_("[-- BEGIN PGP PUBLIC KEY BLOCK --]\n"), s);
	else
	  state_attach_puts (_("[-- BEGIN PGP SIGNED MESSAGE --]\n\n"), s);
      }

      if (clearsign)
      {
	rewind (tmpfp);
	if (tmpfp) 
	  copy_clearsigned (tmpfp, s, body_charset);
      }
      else if (pgpout)
      {
	FGETCONV *fc;
	int c;
	rewind (pgpout);
	fc = fgetconv_open (pgpout, "utf-8", Charset, 0);
	while ((c = fgetconv (fc)) != EOF)
	{
	  state_putc (c, s);
	  if (c == '\n' && s->prefix)
	    state_puts (s->prefix, s);
	}
	fgetconv_close (&fc);
      }

      if (s->flags & M_DISPLAY)
      {
	state_putc ('\n', s);
	if (needpass)
	  state_attach_puts (_("[-- END PGP MESSAGE --]\n"), s);
	else if (pgp_keyblock)
	  state_attach_puts (_("[-- END PGP PUBLIC KEY BLOCK --]\n"), s);
	else
	  state_attach_puts (_("[-- END PGP SIGNED MESSAGE --]\n"), s);
      }

      if (tmpfp)
      {
	safe_fclose (&tmpfp);
	mutt_unlink (tmpfname);
      }
      if (pgpout)
      {
	safe_fclose (&pgpout);
	mutt_unlink (outfile);
      }
    }
    else
    {
      /* XXX - we may wish to recode here */
      if (s->prefix)
	state_puts (s->prefix, s);
      state_puts (buf, s);
    }
  }

  m->goodsig = (maybe_goodsig && have_any_sigs);

  if (needpass == -1)
  {
    state_attach_puts (_("[-- Error: could not find beginning of PGP message! --]\n\n"), s);
    return;
  }
}


int pgp_decrypt_mime (FILE *fpin, FILE **fpout, BODY *b, BODY **cur)
{
  char tempfile[_POSIX_PATH_MAX];
  STATE s;
  BODY *p = b;
  
  if(!mutt_is_multipart_encrypted(b))
    return -1;

  if(!b->parts || !b->parts->next)
    return -1;
  
  b = b->parts->next;
  
  memset (&s, 0, sizeof (s));
  s.fpin = fpin;
  mutt_mktemp (tempfile);
  if ((*fpout = safe_fopen (tempfile, "w+")) == NULL)
  {
    mutt_perror (tempfile);
    return (-1);
  }
  unlink (tempfile);

  *cur = pgp_decrypt_part (b, &s, *fpout, p);

  rewind (*fpout);
  return (0);
}


void pgp_encrypted_handler (BODY *a, STATE *s)
{
  char tempfile[_POSIX_PATH_MAX];
  FILE *fpout, *fpin;
  BODY *tattach;
  BODY *p = a;
  
  a = a->parts;
  if (!a || a->type != TYPEAPPLICATION || !a->subtype || 
      ascii_strcasecmp ("pgp-encrypted", a->subtype) != 0 ||
      !a->next || a->next->type != TYPEAPPLICATION || !a->next->subtype ||
      ascii_strcasecmp ("octet-stream", a->next->subtype) != 0)
  {
    if (s->flags & M_DISPLAY)
      state_attach_puts (_("[-- Error: malformed PGP/MIME message! --]\n\n"), s);
    return;
  }

  /*
   * Move forward to the application/pgp-encrypted body.
   */
  a = a->next;

  mutt_mktemp (tempfile);
  if ((fpout = safe_fopen (tempfile, "w+")) == NULL)
  {
    if (s->flags & M_DISPLAY)
      state_attach_puts (_("[-- Error: could not create temporary file! --]\n"), s);
    return;
  }

  if (s->flags & M_DISPLAY)
    state_attach_puts (_("[-- Begin encryption information --]\n"), s);

  tattach = pgp_decrypt_part (a, s, fpout, p);
  if (tattach)
    {
      if (s->flags & M_DISPLAY)
        state_attach_puts (_("[-- The following data is "
                             "PGP/MIME encrypted --]\n\n"), s);
      
      fpin = s->fpin;
      s->fpin = fpout;
      mutt_body_handler (tattach, s);
      s->fpin = fpin;
      
      /* 
       * if a multipart/signed is the _only_ sub-part of a
       * multipart/encrypted, cache signature verification
       * status.
       */
    
      if (mutt_is_multipart_signed (tattach) && !tattach->next)
        p->goodsig |= tattach->goodsig;
    
      if (s->flags & M_DISPLAY)
        {
          state_puts ("\n", s);
          state_attach_puts (_("[-- End of PGP/MIME encrypted data --]\n"), s);
        }

      mutt_free_body (&tattach);
    }
  
  fclose (fpout);
  mutt_unlink(tempfile);
}


/*
  This handles application/pkcs7-mime which can either be a signed
  or an encrypted message.
*/
static BODY *smime_handle_entity (BODY *m, STATE *s, FILE *outFile)
{
#if 0
  int len=0;
  int c;
  long last_pos;
  char buf[HUGE_STRING];
  char outfile[_POSIX_PATH_MAX], errfile[_POSIX_PATH_MAX];
  char tmpfname[_POSIX_PATH_MAX];
  char tmptmpfname[_POSIX_PATH_MAX];
  FILE *smimeout = NULL, *smimein=NULL, *smimeerr=NULL;
  FILE *tmpfp=NULL, *tmpfp_buffer=NULL, *fpout=NULL;
  struct stat info;
  BODY *p=NULL;
  pid_t thepid=-1;
  unsigned int type = mutt_is_application_smime (m);

  if (!(type & APPLICATION_SMIME))
    return NULL;

  mutt_mktemp (outfile);
  if ((smimeout = safe_fopen (outfile, "w+")) == NULL)
  {
    mutt_perror (outfile);
    return NULL;
  }
  
  mutt_mktemp (errfile);
  if ((smimeerr = safe_fopen (errfile, "w+")) == NULL)
  {
    mutt_perror (errfile);
    fclose (smimeout); smimeout = NULL;
    return NULL;
  }
  mutt_unlink (errfile);

  
  mutt_mktemp (tmpfname);
  if ((tmpfp = safe_fopen (tmpfname, "w+")) == NULL)
  {
    mutt_perror (tmpfname);
    fclose (smimeout); smimeout = NULL;
    fclose (smimeerr); smimeerr = NULL;
    return NULL;
  }

  fseek (s->fpin, m->offset, 0);
  last_pos = m->offset;

  mutt_copy_bytes (s->fpin, tmpfp,  m->length);

  fflush (tmpfp);
  fclose (tmpfp);

  if ((type & ENCRYPT) &&
      (thepid = smime_invoke_decrypt (&smimein, NULL, NULL, -1,
				      fileno (smimeout),  fileno (smimeerr), tmpfname)) == -1)
  {
    fclose (smimeout); smimeout = NULL;
    mutt_unlink (tmpfname);
    state_attach_puts (_("[-- Error: unable to create OpenSSL subprocess! --]\n"), s);
    return NULL;
  }
  else if ((type & SIGNOPAQUE) &&
	   (thepid = smime_invoke_verify (&smimein, NULL, NULL, -1,
					  fileno (smimeout), fileno (smimeerr), NULL,
					  tmpfname, SIGNOPAQUE)) == -1)
  {
    fclose (smimeout); smimeout = NULL;
    mutt_unlink (tmpfname);
    state_attach_puts (_("[-- Error: unable to create OpenSSL subprocess! --]\n"), s);
    return NULL;
  }

  
  if (type & ENCRYPT)
  {
    if (!smime_valid_passphrase ())
      smime_void_passphrase ();
    fputs (SmimePass, smimein);
    fputc ('\n', smimein);
  }

  fclose (smimein);
	
  mutt_wait_filter (thepid);
  mutt_unlink (tmpfname);
  

  if (s->flags & M_DISPLAY)
  {
    rewind (smimeerr);
    
    if ((c = fgetc (smimeerr)) != EOF)
    {
      ungetc (c, smimeerr);
      
      crypt_current_time (s, "GnuPG");
      mutt_copy_stream (smimeerr, s->fpout);
      state_attach_puts (_("[-- End of GnuPG output --]\n\n"), s);
    }
    
    if (type & ENCRYPT)
      state_attach_puts (_("[-- The following data is S/MIME"
                           " encrypted --]\n"), s);
    else
      state_attach_puts (_("[-- The following data is S/MIME signed --]\n"), s);
  }

  if (smimeout)
  {
    fflush (smimeout);
    rewind (smimeout);
    
    if (outFile) fpout = outFile;
    else
    {
      mutt_mktemp (tmptmpfname);
      if ((fpout = safe_fopen (tmptmpfname, "w+")) == NULL)
      {
	mutt_perror(tmptmpfname);
	fclose (smimeout); smimeout = NULL;
	return NULL;
      }
    }
    while (fgets (buf, sizeof (buf) - 1, smimeout) != NULL)
    {
      len = mutt_strlen (buf);
      if (len > 1 && buf[len - 2] == '\r')
	strcpy (buf + len - 2, "\n");
      fputs (buf, fpout);
    }
    fflush (fpout);
    rewind (fpout); 


    if ((p = mutt_read_mime_header (fpout, 0)) != NULL)
    {
      fstat (fileno (fpout), &info);
      p->length = info.st_size - p->offset;
	  
      mutt_parse_part (fpout, p);
      if (s->fpout)
      {
	rewind (fpout);
	tmpfp_buffer = s->fpin;
	s->fpin = fpout;
	mutt_body_handler (p, s);
	s->fpin = tmpfp_buffer;
      }
      
    }
    fclose (smimeout);
    smimeout = NULL;
    mutt_unlink (outfile);

    if (!outFile)
    {
      fclose (fpout);
      mutt_unlink (tmptmpfname);
    }
    fpout = NULL;
  }

  if (s->flags & M_DISPLAY)
  {
    if (type & ENCRYPT)
      state_attach_puts (_("\n[-- End of S/MIME encrypted data. --]\n"), s);
    else
      state_attach_puts (_("\n[-- End of S/MIME signed data. --]\n"), s);
  }

  if (type & SIGNOPAQUE)
  {
    char *line = NULL;
    int lineno = 0;
    size_t linelen;
    
    rewind (smimeerr);
    
    line = mutt_read_line (line, &linelen, smimeerr, &lineno);
    if (linelen && !mutt_strcasecmp (line, "verification successful"))
      m->goodsig = 1;
    safe_free ((void **) &line);
  }
  else {
    m->goodsig = p->goodsig;
    m->badsig = p->badsig;
  }
  fclose (smimeerr);

  return (p);
#endif
  return NULL;
}


void smime_application_smime_handler (BODY *m, STATE *s)
{
  return; /* FIXME */
}


int smime_decrypt_mime (FILE *fpin, FILE **fpout, BODY *b, BODY **cur)
{
  char tempfile[_POSIX_PATH_MAX];
  STATE s;
  long tmpoffset = b->offset;
  size_t tmplength = b->length;
  int origType = b->type;
  FILE *tmpfp=NULL;

  if (!mutt_is_application_smime (b))
    return -1;

  if (b->parts)
    return -1;
  
  memset (&s, 0, sizeof (s));
  s.fpin = fpin;
  fseek (s.fpin, b->offset, 0);

  mutt_mktemp (tempfile);
  if ((tmpfp = safe_fopen (tempfile, "w+")) == NULL)
  {
    mutt_perror (tempfile);
    return (-1);
  }

  mutt_unlink (tempfile);
  s.fpout = tmpfp;
  mutt_decode_attachment (b, &s);
  fflush (tmpfp);
  b->length = ftell (s.fpout);
  b->offset = 0;
  rewind (tmpfp);
  s.fpin = tmpfp;
  s.fpout = 0;

  mutt_mktemp (tempfile);
  if ((*fpout = safe_fopen (tempfile, "w+")) == NULL)
  {
    mutt_perror (tempfile);
    return (-1);
  }
  mutt_unlink (tempfile);

  *cur = smime_handle_entity (b, &s, *fpout);
  (*cur)->goodsig = b->goodsig;
  (*cur)->badsig = b->badsig;
  b->type = origType;
  b->length = tmplength;
  b->offset = tmpoffset;
  fclose (tmpfp);

  rewind (*fpout);
  return (0);
}



int pgp_check_traditional (FILE *fp, BODY *b, int tagged_only)
{
  int rv = 0;
#if 0
  int r;
  for (; b; b = b->next)
  {
    if (is_multipart (b))
      rv = pgp_check_traditional (fp, b->parts, tagged_only) || rv;
    else if (b->type == TYPETEXT)
    {
      if ((r = mutt_is_application_pgp (b)))
	rv = rv || r;
      else
	rv = pgp_check_traditional_one_body (fp, b, tagged_only) || rv;
    }
  }

#endif
  return rv;
}

#endif /*HAVE_GPGME*/
